#include <Arduino.h>
#include <WiFi.h>
#include <Firebase_ESP_Client.h>
#include <Wire.h>
#include <BH1750.h>
#include <sSense-CCS811.h>
#include <sSense-HDC2010.h>
#define CUSTOM_SETTINGS       // usefull for Dabble library
#define INCLUDE_SENSOR_MODULE // usefull for Dabble library
#include <DabbleESP32.h>
#include "config.h"

//------------------------ GLOBAL VARIABLES --------------------------

// Sensors variables
CCS811 ssenseCCS811;

HDC2010 ssenseHDC2010(HDC2010_I2C_ADDR);

BH1750 lightMeter;

// Firebase objects
FirebaseData stream;
FirebaseData fbdo;
FirebaseAuth auth;
FirebaseConfig config;

// Database paths
String userDataPath;
String recordingStatePath;
String confirmationPath;
String recordsNumberPath;
String recordsPath;

// database states
bool recordingState = false;
int lastRecordId;
int lastValueId = -1;

// for time intervalls
unsigned long sendDataPrevMillis = 0;

//-----------------Helpers Declarations-----------------

//----init functions---------------------

void init_WiFi();
void init_BH1750();
void init_Firebase();
void initUserPaths(); // init the database paths for the specific logged in user
void init_ssenseCCS811();
void init_ssenseHDC2010();

//---------Firebase-dedicated functions--------------------
// Listeners

// Callback function that runs on database changes
void recordingStateListener(FirebaseStream data);
void streamTimeoutCallback(bool timeout);
// Functions related to the Application's database structure
int getNumberOfRecords(FirebaseData *fbdo);
bool incrementNumberOfRecords(FirebaseData * fbdo);
int getRecordSize(FirebaseData *fbdo, int recordId);
bool incrementRecordSize(FirebaseData *fbdo, int recordId);

bool sendData(FirebaseData *fbdo, float light, float sound, float co2, float temperature, float humidity, float global);
bool getAndSendDataFromSensors();


//-------------------Others---------------------
float computeGlobalGrade(float light, float sound, float co2, float  temperature, float humidity);

//------------------------------------------------------


void setup()
{
  DebugPort.begin(SERIAL_SPEED);

  init_BH1750();
  init_ssenseCCS811();
  init_ssenseHDC2010();

  Dabble.begin(BLUETOOTH_DEVICE_NAME); // init Dabble

  init_WiFi();
  init_Firebase();

  delay(2000);
}

void loop()
{
  if (Firebase.isTokenExpired())
  {
    Firebase.refreshToken(&config);
    DebugPort.println("Refresh token");
  }

  if (Firebase.ready() && (millis() - sendDataPrevMillis > TIME_INTERVAL || sendDataPrevMillis == 0))
  {

    sendDataPrevMillis = millis();
    if (recordingState)
    {
      DebugPort.println("Sending data");
      if (!getAndSendDataFromSensors())
      {
        DebugPort.println("Error while sending data from sensors");
      }
    }
  }
}

//-----Helper functions defintions---------------------

// Initialize WiFi
void init_WiFi()
{
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
  DebugPort.print("Connecting to WiFi ..");
  while (WiFi.status() != WL_CONNECTED)
  {
    DebugPort.print('.');
    delay(1000);
  }
  DebugPort.println(WiFi.localIP());
  DebugPort.println();
}

void init_ssenseCCS811()
{
  DebugPort.println("s-Sense CCS811 I2C sensor.");
  if (!ssenseCCS811.begin(uint8_t(I2C_CCS811_ADDRESS), uint8_t(CCS811_WAKE_PIN), driveMode_1sec))
    DebugPort.println("Initialization failed.");
}

void init_ssenseHDC2010()
{
  // Initialize HDC2010(THS) I2C communication
  ssenseHDC2010.begin();

  // Begin with a HDC2010(THS) reset
  ssenseHDC2010.reset();

  ssenseHDC2010.setTemperatureOffset(0b11010111); //-6.64 degrees Celsius required
  // Configure Measurements
  ssenseHDC2010.setMeasurementMode(TEMP_AND_HUMID); // Set measurements to temperature and humidity
  ssenseHDC2010.setRate(ONE_HZ);                    // Set measurement frequency to 1 Hz
  ssenseHDC2010.setTempRes(FOURTEEN_BIT);
  ssenseHDC2010.setHumidRes(FOURTEEN_BIT);

  delay(1000);

  // begin HDC2010 sensor measuring
  ssenseHDC2010.triggerMeasurement();
}

void init_BH1750()
{

  DebugPort.println("BH1750 I2C sensor.");
  Wire.begin();
  lightMeter.begin();
}

void initUserPaths()
{
  userDataPath = USERS_DATA_BASE_PATH + String(auth.token.uid.c_str()) + "/";
  recordingStatePath = userDataPath + RECORDING_STATE_BASE_PATH;
  confirmationPath = userDataPath + CONFIRMATION_BASE_PATH;
  recordsNumberPath = userDataPath + RECORDS_NUMBER_BASE_PATH;
  recordsPath = userDataPath + RECORDS_BASE_PATH;
}

void init_Firebase()
{
  // Assign the api key (required)
  config.api_key = API_KEY;

  // Assign the user sign in credentials
  auth.user.email = USER_EMAIL;
  auth.user.password = USER_PASSWORD;

  // Assign the RTDB URL (required)
  config.database_url = DATABASE_URL;

  Firebase.reconnectWiFi(true);

  // Assign the maximum retry of token generation
  config.max_token_generation_retry = 5;

  // Initialize the library with the Firebase authen and config
  Firebase.begin(&config, &auth);
  if (Firebase.ready())
  {
    initUserPaths();

    lastRecordId = getNumberOfRecords(&fbdo) - 1;
    lastValueId = getRecordSize(&fbdo, lastRecordId) - 1;

    // Streaming (whenever data changes on a path)
    // Begin stream on a database path --> .../recordingInProgress
    if (!Firebase.RTDB.beginStream(&stream, recordingStatePath.c_str()))
      DebugPort.printf("stream begin error, %s\n\n", stream.errorReason().c_str());

    // Assign a calback function to run when it detects changes on the database
    Firebase.RTDB.setStreamCallback(&stream, recordingStateListener, streamTimeoutCallback);
  }
  else
  {
    DebugPort.println("Error: Firebase not ready");
  }
}

bool getAndSendDataFromSensors()
{
  float temperature = ssenseHDC2010.readTemp();
  float humidity = ssenseHDC2010.readHumidity();
  DebugPort.print("Temperature ");
  DebugPort.print(temperature);
  DebugPort.println(" °C");
  DebugPort.print("Humidity : ");
  DebugPort.print(humidity);
  DebugPort.println(" %");

  ssenseCCS811.wakeup(); // wake the sensor up
  // set temperature and humidity to better compensation
  ssenseCCS811.setEnvironmentalData(temperature, humidity);

  float co2;
  if (ssenseCCS811.checkDataAndUpdate())
  {
    co2 = ssenseCCS811.getCO2();
    DebugPort.print("CO2 : ");
    DebugPort.print(co2);
    DebugPort.println(" ppm");
  }
  else if (ssenseCCS811.checkForError())
  {

    ssenseCCS811.printError();
  }
  ssenseCCS811.sleep(); // turn the sensor into sleep mode

  float light = lightMeter.readLightLevel();
  DebugPort.print("LIGHT : ");
  DebugPort.print(light);
  DebugPort.println(" lux");

  Dabble.processInput(); // refresh data obtained from smartphone
  float sound = Sensor.getSoundDecibels() + SOUND_OFFSET;

  // human hearing threshold is 0
  sound = sound < 0 ? 0 : sound;

  DebugPort.print("SOUND : ");
  DebugPort.print(sound);
  DebugPort.println(" dB");

  float global = computeGlobalGrade(light, sound, co2, temperature, humidity);

  DebugPort.print("GLOBAL : ");
  DebugPort.print(global*100);
  DebugPort.println("/100");
  return sendData(&fbdo, light, sound, co2, temperature, humidity, global);
}

float computeGlobalGrade(float light, float sound, float co2, float temperature, float humidity)
{
  float globalGrade = 0;

  // light grading
  if (light <= 5)
  {
    globalGrade += 0.2;
  }
  else if (light <= 15)
  {
    globalGrade += 0.15;
  }
  else if (light <= 40)
  {
    globalGrade += 0.10;
  }
  else if (light <= 100)
  {
    globalGrade += 0.05;
  }
  else if (light <= 200)
  {
    globalGrade += 0;
  }
  else
  {
    globalGrade -= 0.1;
  }

  // sound grading
  if (sound <= 25)
  {
    globalGrade += 0.2;
  }
  else if (sound <= 30)
  {
    globalGrade += 0.15;
  }
  else if (sound <= 40)
  {
    globalGrade += 0.10;
  }
  else if (sound <= 55)
  {
    globalGrade += 0.05;
  }
  else if (sound <= 75)
  {
    globalGrade += 0;
  }
  else
  {
    globalGrade -= 0.1;
  }

  // co2 grading
  if (co2 <= 400)
  {
    globalGrade += 0.2;
  }
  else if (co2 <= 700)
  {
    globalGrade += 0.15;
  }
  else if (co2 <= 1000)
  {
    globalGrade += 0.15;
  }
  else if (co2 <= 2000)
  {
    globalGrade += 0.05;
  }
  else if (co2 <= 5000)
  {
    globalGrade += 0;
  }
  else
  {
    globalGrade -= 0.1;
  }

  // temperature grading
  if (temperature >= 17 && temperature <= 18)
  {
    globalGrade += 0.2;
  }
  else if (temperature >= 16 && temperature <= 20)
  {
    globalGrade += 0.15;
  }
  else if (temperature >= 14 && temperature <= 25)
  {
    globalGrade += 0.1;
  }
  else if (temperature >= 12 && temperature <= 28)
  {
    globalGrade += 0.05;
  }
  else if (temperature >= 10 && temperature <= 30)
  {
    globalGrade += 0;
  }
  else
  {
    globalGrade -= 0.1;
  }

  // humidity
  if (humidity >= 40 && humidity <= 60)
  {
    globalGrade += 0.2;
  }
  else if (humidity >= 35 && humidity <= 65)
  {
    globalGrade += 0.15;
  }
  else if (humidity >= 30 && humidity <= 70)
  {
    globalGrade += 0.1;
  }
  else if (humidity >= 20 && humidity <= 80)
  {
    globalGrade += 0.05;
  }
  else if (humidity >= 10 && humidity <= 90)
  {
    globalGrade += 0;
  }
  else
  {
    globalGrade -= 0.1;
  }

  return globalGrade;
}

bool sendData(FirebaseData *fbdo, float light, float sound, float co2, float temp, float hum, float global)
{
  if (!Firebase.ready())
  {
    DebugPort.println("Firebase not ready. Cannot send data.");
    return false;
  }
  lastValueId++;
  String path = String(recordsPath) + String(lastRecordId) + "/" + String(lastValueId) + "/";

  FirebaseJson json;
  json.add(LIGHT_FIELD_NAME, light);
  json.add(SOUND_FIELD_NAME, sound);
  json.add(CO2_FIELD_NAME, co2);
  json.add(TEMPERATURE_FIELD_NAME, temp);
  json.add(HUMIDITY_FIELD_NAME, hum);
  json.add(GLOBAL_FIELD_NAME, global);

  if (Firebase.RTDB.setJSON(fbdo, path.c_str(), &json) && Firebase.RTDB.setTimestamp(fbdo, (path + TIMESTAMP_FIELD_NAME).c_str()))
  {
    incrementRecordSize(fbdo, lastRecordId);
    DebugPort.println("Data sent successfully");
    return true;
  }
  else
  {
    DebugPort.println("Failed to send data");
    DebugPort.println("Error: " + stream.errorReason());
    return false;
  }
}

int getNumberOfRecords(FirebaseData *fbdo)
{
  if (Firebase.RTDB.getInt(fbdo, recordsNumberPath))
  {
    return fbdo->intData();
  }
  else
  {
    return 0;
  }
}

bool incrementNumberOfRecords(FirebaseData *fbdo)
{
  int numberOfrecords = getNumberOfRecords(fbdo);
  lastRecordId = numberOfrecords;
  lastValueId = -1;
  return Firebase.RTDB.setInt(fbdo, recordsNumberPath, numberOfrecords + 1);
}

int getRecordSize(FirebaseData *fbdo, int recordId)
{
  if (Firebase.RTDB.getInt(fbdo, recordsPath + recordId + "/size"))
  {
    return fbdo->intData();
  }
  else
  {
    return 0;
  }
}

bool incrementRecordSize(FirebaseData *fbdo, int recordId)
{
  int recordSize = getRecordSize(fbdo, recordId);
  lastValueId = recordSize;
  return Firebase.RTDB.setInt(fbdo, recordsPath + recordId + "/size", recordSize + 1);
}

void recordingStateListener(FirebaseStream data)
{

  // change the recording state
  bool previousRecordingState = recordingState;
  recordingState = data.boolData();
  DebugPort.print("STATE: ");
  DebugPort.println(recordingState);
  if (recordingState && !(recordingState & previousRecordingState) && Firebase.RTDB.getInt(&stream, confirmationPath))
  {
    sendDataPrevMillis = -TIME_INTERVAL;
    if (stream.intData() == -1)
    {
      if (Firebase.RTDB.setTimestamp(&stream, confirmationPath))
      {
        incrementNumberOfRecords(&stream);
        DebugPort.println("Confirmed");
      }
      else
      {
        DebugPort.println("FAILED");
        DebugPort.println("REASON: " + stream.errorReason());
      }
    }
  }
}

void streamTimeoutCallback(bool timeout)
{
  if (timeout)
    DebugPort.println("stream timeout, resuming...\n");
  if (!stream.httpConnected())
    DebugPort.printf("error code: %d, reason: %s\n\n", stream.httpCode(), stream.errorReason().c_str());
}

//------------------------------------------------------